#include <linux/limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif


#define RESTART 10
#define SECRETSZ 32
#define CMDSZ 9
#define MAXCMDS ('z' - 'a' + 1)
char secret[SECRETSZ];

int readsecret();
int readn(char *p, int n);

typedef void (*cmd_t)(char *arg);

void cmd_restart(char *arg);
void cmd_secret(char *arg);
void cmd_echo(char *arg);
void cmd_null(char *arg);
cmd_t cmds[MAXCMDS] = {
	['r' - 'a'] = cmd_restart,
	['s' - 'a'] = cmd_secret,
	['e' - 'a'] = cmd_echo,
};

void child()
{
	char cmd[CMDSZ];
	int i;

	readsecret();

	for (i = 0; i < MAXCMDS; i++) {
		if (cmds[i] == NULL)
			cmds[i] = cmd_null;
	}

	while (readn(cmd, CMDSZ) > 0) {
		char c = cmd[0];
		if (c < 'a' || c > 'z')
			continue;
		cmds[c - 'a'](cmd + 1);
	}
}

void cmd_restart(char *arg)
{
	exit(RESTART);
}

void cmd_null(char *arg)
{
}

// we use printf() for convenience, but we use ASLR+PIE and
// we randomly pad our stack
void cmd_secret(char *arg)
{
	char dummy[PADDING];
	printf("%s", ((char**)arg)[0]);
	cmds['s'] = cmd_restart;
}

void cmd_echo(char *arg)
{
	char dummy[PADDING];
	printf(arg);
}

/* NO INTENTIONAL ERRORS STARTING FROM THIS POINT */

int readsecret()
{
	int fd, i;
	char c, *dst;

	fd = open("flag.txt", O_RDONLY);
	if (fd < 0)
		return -1;

	for (i = 0, dst = secret; i < SECRETSZ; i++, dst++) {
		if (read(fd, dst, 1) < 0)
			return -1;
	}
	close(fd);
}

int readn(char *p, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (read(0, p, 1) <= 0)
			return -1;
		p++;
	}
	return 1;
}

int writen(char *p, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (write(1, p, 1) != 1)
			return -1;
		p++;
	}
	return 0;
}


void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
			if (WEXITSTATUS(status) == RESTART)
				exit(RESTART);
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
			exit(RESTART);
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;
	char * volatile dummy;

	srand(time(NULL));

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			child();
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}
